from collections import defaultdict

from cml.shared.errors import SelectionFailure, WinnerSelectionFailure
from krippendorff import alpha


__all__ = (
    "FeatureSelector",
)


def determine_winner(reliability_dict):
    if not reliability_dict:
        raise WinnerSelectionFailure()

    biggest_reliabllity = max(reliability_dict.keys())
    winners = reliability_dict[biggest_reliabllity]

    winner_model = None
    winner_learnblock = None
    max_domain_length = float("inf")
    for model, lb in winners:
        if len(model.pre_image_features) < max_domain_length:
            max_domain_length = len(model.pre_image_features)
            winner_model = model
            winner_learnblock = lb

    return winner_model, winner_learnblock


    #if not isinstance(reliability_list, list):
    #    reliability_list = [reliability_list]

    #reliabilities_to_pair = defaultdict(list)
    #for reliability_dict in reliability_list:
    #    for reliability, pair in reliability_dict.items():
    #        reliabilities_to_pair[reliability].extend(pair)

    #try:
    #    sorted_reliabilities = sorted(reliabilities_to_pair.keys(), reverse=False)
    #    biggest_reliabilities = reliabilities_to_pair[sorted_reliabilities.pop()]
    #    winner = None
    #    winner_lb = None
    #    min_domain = float("inf")
    #    for pair in biggest_reliabilities:
    #        if len(pair.model.pre_image_features) < min_domain:
    #            min_domain = len(pair.model.pre_image_features)
    #            winner, winner_lb = pair.model, pair.lb

    #    return winner, winner_lb

    #except IndexError:
    #    raise WinnerSelectionFailure()


class FeatureSelector:
    def __init__(self, filter_method, embedded_method, settings):
        self.filter_method = filter_method
        self.embedded_method = embedded_method
        self.settings = settings

    def select(self, lb):
        while 1:
            if self.reduce_constraint(lb):
                past_n_feature = lb.n_columns(effective=True)
                self.reduce(lb)
                if self.none_reducible_valid_learnblock(past_n_feature, lb):
                    return lb
                elif self.none_reducible_invalid_learnblock(past_n_feature, lb):
                    raise SelectionFailure()
                else:
                    # reduce further
                    continue
            else:
                return lb

    def reduce(self, learnblock):
        if self._filter_method_criteria(learnblock.n_columns(effective=True),
                                        learnblock.n_rows()):
            remove_features = self.filtering(learnblock)
            learnblock.drop_columns(remove_features, by_index=True)
        else:
            remove_features = self.embedding(learnblock)
            learnblock.drop_columns_by_index(remove_features)

    def reduce_constraint(self, lb):
        return self._to_many_features(lb.n_columns(effective=True)) or \
            self.settings.max_model_reduction

    def none_reducible_valid_learnblock(self, old_lb_features, lb):
        return old_lb_features == lb.n_columns(effective=True) and not \
            self._to_many_features(lb.n_columns(effective=True))

    def none_reducible_invalid_learnblock(self, past_n_features, lb):
        return past_n_features == lb.n_columns(effective=True) and \
               self._to_many_features(lb.n_columns(effective=True))

    def filtering(self, learnblock):
        trained_model = self.filter_method.train(learnblock)
        return trained_model.reduce(learnblock)

    def embedding(self, learnblock):
        trained_model = self.embedded_method.train(learnblock)
        return trained_model.reduce()

    def _to_many_features(self, learn_cols):
        return learn_cols > self.settings.max_features

    def _filter_method_criteria(self, learn_cols, samples_count):
        return learn_cols > self.settings.max_filter_x or \
                samples_count > self.settings.max_filter_y
